import Data.Map as Map
import Data.List as List
import System.Random
import Control.Parallel.Strategies
import System.Exit(die)
import System.Environment(getArgs, getProgName)

generateRandomCycle :: Int -> Int -> IO([Int])
generateRandomCycle sz n = do
    g <- newStdGen
    return (List.take n . nub $ (randomRs (0, sz) g :: [Int]))

generateRandomValues :: Int -> Int -> Int -> IO([Int])
generateRandomValues 0 _ _= return []
generateRandomValues n l h = do
    r <- randomRIO (l, h)
    rs <- generateRandomValues ((-) n 1) l h
    return (r:rs)

getOldWeight :: (Num a, Ord k) => [k] -> Map k (Map k a) -> [a]
getOldWeight c g
    | (length c) < 2 = []
    | otherwise = fromJust (weight g (head c) (head tc)) : getOldWeight tc g
        where
            tc = tail c

setWeightToMap :: (Ord a1, Num a2) =>
                        Map a1 (Map a1 a2) -> [a2] -> [a1] -> Map a1 (Map a1 a2)
setWeightToMap g w c
                | (length c) < 2 = g
                | otherwise = setWeightToMap (insertGraph (insertGraph g hc htc hw) htc hc ((-1) * hw)) (tail w) tc
                    where
                        tc = tail c
                        htc = head tc
                        hc = head c
                        hw = head w

generateNewWeightForNegCycle :: Ord k =>
                                      [k] -> Map k (Map k Int) -> IO (Map k (Map k Int), Int)
generateNewWeightForNegCycle cyc graphMap = do
    let oldWeight = getOldWeight cyc graphMap
    addWeights  <- generateRandomValues ((-) (length cyc) 1) 1 100
    let newWeights = zipWith (+) oldWeight addWeights
    return ((setWeightToMap graphMap newWeights cyc), (List.foldr (+) 0 addWeights))

weight :: (Ord k1, Ord k2) =>
                Map k1 (Map k2 b) -> k1 -> k2 -> Maybe b
weight graph i j = do
    rowMap <- Map.lookup i graph
    Map.lookup j rowMap

insertGraph :: (Ord k1, Ord k2) =>
                     Map k1 (Map k2 a) -> k1 -> k2 -> a -> Map k1 (Map k2 a)
insertGraph graph i j w = Map.insert i (Map.insert j w (fromJust2 (Map.lookup i graph))) graph

initGraphMap :: (Num a, Ord k, Enum k, Num k) =>
                      k -> Map k (Map k a)
initGraphMap sz = Map.fromList $ Prelude.map f [0..((-) sz 1)]
    where
        f i = (i, Map.fromList $ zip [0..((-) sz 1)] (repeat 0))

fromJust :: Num p => Maybe p -> p
fromJust (Just a) = a
fromJust Nothing = 0

fromJust2 :: Ord k => Maybe (Map k a) -> Map k a
fromJust2 (Just a) = a
fromJust2 Nothing = Map.fromList []

getVertices :: Map b1 b2 -> [b1]
getVertices m = Prelude.map (\(i, _) -> i) (Map.toList m)

fillGraphMap :: (Ord b, Num p, Num b) =>
                      Map b (Map b p) -> Map b p -> Map b (Map b p)
fillGraphMap graph initialValueMap = List.foldl' mapRowFunc graph (getVertices graph)
    where
        mapRowFunc graph1 rowI = Map.insert rowI (Map.mapWithKey mapValFunc (fromJust2 (Map.lookup rowI graph))) graph1
            where
                mapValFunc colJ _
                    | rowI == colJ = 0
                    | rowI == 0 = fromJust (Map.lookup colJ initialValueMap)
                    | rowI > colJ = ((*) (-1) (fromJust (weight graph1 colJ rowI)))
                    | otherwise = ((-) (fromJust (weight graph1 0 colJ)) (fromJust (weight graph1 0 rowI)))

main :: IO ()
main = do
    args <- getArgs
    case args of
        [s, l, filename] -> do
            let sz = (read s :: Int)
            let negCycleLen = (read l :: Int)
            let emptyMap = initGraphMap sz
            randomValues <- generateRandomValues ((-) sz 1) (-100) 100
            let initialValues = Map.fromList (zip [0..] (0 : randomValues))
            let graphMap = fillGraphMap emptyMap initialValues
            randomCycle <- generateRandomCycle sz negCycleLen
            (finalNegGraphMap, extraWeight) <- generateNewWeightForNegCycle randomCycle graphMap
            let f gMap = (\(_, rowMap) -> (intercalate "," . Prelude.map show) (Prelude.map (\(_, x) -> x) (Map.toList rowMap))) gMap
            writeFile (filename ++ "-graph.txt") $ intercalate "\n" ((parMap rdeepseq f (Map.toList finalNegGraphMap)) :: [[Char]])
            writeFile (filename ++ "-cycle.txt") $ intercalate "," $ Prelude.map show randomCycle
            writeFile (filename ++ "-weight.txt") $ show extraWeight
        _ -> do
            pn <- getProgName
            die $ "Usage: " ++ pn ++ " <size> <negCycleLen> <filename>"

